/* ckbd_drv.c
 * Interface to Cherry Keyboards
 */

/* The necessary header files */

/* Standard in kernel modules */
#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/module.h>   /* Specifically, a module */

#include <asm/uaccess.h>  /* for get_user and put_user */

#include <linux/sched.h>

/* The character device definitions are here */
#include <linux/fs.h>


/* Our own definitions */
#include "kbd_devs.h"

/* to output some debug messages */
//#define MY_DEBUG

/* Defines */

#define STAT_LEN 8

#define LEFT_CTRL 0x01
#define LEFT_SHFT 0x02
#define LEFT_ALT  0x04
#define LEFT_WIN  0x08
#define RIGHT_CTRL 0x10
#define RIGHT_SHFT 0x20
#define RIGHT_ALT  0x40
#define RIGHT_WIN  0x80

/* Device Declarations ******************************** */
/* for timeout timer */
struct timer_list jiq_timer;

/* The name for our device, as it will appear in
 * /proc/devices */
#define DEVICE_NAME "chyposusbkbd"


/* The maximum length of the message for the device */
#define BUF_LEN 128

/* Is the device open right now? Used to prevent
 * concurent access into the same device */
static int Device_Open = 0;
static int Device_Minor = 0;
static int Device_Major = 0;

static spinlock_t       ckbd_u_lock;
static int ckbd_u_count = 0;
static unsigned int ckbd_u_owner = 0;

static int ckbd_u_stat = 0;
static int ckbd_u_keyboard = 0;

/* The message the device will give when asked */
static char Message[BUF_LEN];

/* scancode history for protokoll separation */
static unsigned char NewKbdData[8];
static unsigned char OldKbdData[8];

/* How far did the process reading the message get?
 * Useful if the message is larger than the size of the
 * buffer we get to fill in device_read. */
static char *Message_Ptr;
static size_t Message_Len = 0;

static unsigned long *kdata;

/* Stat data area */
static struct {
  unsigned char keylock;
  unsigned char scancodenr;
} StatData;
 
static int STAT_len = 0;
static unsigned  char *pStatData;
#define KEYLOCK_STATUS  StatData.keylock
#define KEYBOARD_STATUS  StatData.scancodenr

static char PrefixFlag;

/* Keyboard data area */
static unsigned char KeyboardData[BUF_LEN];
static unsigned char *pKeyboardData;

#define KEYCONF_LEN 1024

/* 256 words poskeywerte 
   256 words poskeywerte fuer prefixkeys
   
   poskeytabelle (x) != 0 bedeutet usb scancode (x) = POSKEY wenn modifier entspechend modifiertabelle (x) gedrueckt sind
   prefixtabelle (x) analog fuer prefixkeys.
   
   
Sonderbehandlungen:
   bei poskeytabelle (0) != 0 sind alle Tasten Poskeys POSKEYS
   modifyertabelle (0) ist prefix modifyer
   prefixkeys(0)  ist prefix code
*/

/* block internal handling of scancodes */

static unsigned char keyconf[(KEYCONF_LEN)];

/* This function is called whenever a process attempts
 * to open the device file */
static int device_open(struct inode *inode,
                       struct file *file)
{
#ifdef MY_DEBUG
  printk("device_open(%p)\n", file);
#endif

  spin_lock(&ckbd_u_lock);
  if(ckbd_u_count &&
     (ckbd_u_owner != current->uid) &&  /* allow user */
     (ckbd_u_owner != current->euid) && /* allow whoever did su */
     !capable(CAP_DAC_OVERRIDE)) {      /* still allow root */
    spin_unlock(&ckbd_u_lock);
    return -EBUSY;      /* -EPERM would confuse the user */
  }

  if(ckbd_u_count == 0) {
    ckbd_u_owner = current->uid;        /* grab it */
  }

  ckbd_u_count++;
  spin_unlock(&ckbd_u_lock);


  /* If this was a process, we would have had to be
   * more careful here, because one process might have
   * checked Device_Open right before the other one
   * tried to increment it. However, we're in the
   * kernel, so we're protected against context switches.
   *
   * This is NOT the right attitude to take, because we
   * might be running on an SMP box, but we'll deal with
   * SMP in a later chapter.
   */

  Device_Open++;
  Device_Minor = MINOR(inode->i_rdev);
#ifdef MY_DEBUG
  printk ("device_open minor = %d\n",Device_Minor);
#endif
  switch(Device_Minor) {
    case STAT_MINOR_NUM:
      ckbd_u_stat++;
      break;
    case KEYBOARD_MINOR_NUM:
      ckbd_u_keyboard++;
      pKeyboardData = &KeyboardData[0];
      KEYBOARD_STATUS = 0;
      break;
  }
  /* Initialize the message */
  Message_Ptr = Message;
  Message_Len = 0;

  return 0;
}


/* This function is called when a process closes the
 * device file. It doesn't have a return value because
 * it cannot fail. Regardless of what else happens, you
 * should always be able to close a device (in 2.0, a 2.2
 * device file could be impossible to close). */
static int device_release(struct inode *inode,
                          struct file *file)
{
#ifdef MY_DEBUG
  printk ("device_release(%p,%p)\n", inode, file);
#endif
  Device_Minor = MINOR(inode->i_rdev);
#ifdef MY_DEBUG
  printk ("device_release minor = %d\n",Device_Minor);
#endif
  switch(Device_Minor) {
    case STAT_MINOR_NUM:
      ckbd_u_stat--;
      break;
    case KEYBOARD_MINOR_NUM:
      ckbd_u_keyboard--;
      break;
  }

  /* We're now ready for our next caller */
  Device_Open --;
  ckbd_u_count--;

  return 0;
}



/* This function is called whenever a process which
 * has already opened the device file attempts to
 * read from it. */

static ssize_t device_read(
    struct file *file,
    char *buffer,   /* The buffer to fill with the data */
    size_t length,  /* The length of the buffer */
    loff_t *offset) /* offset to the file */
{
  int bytes_read = 0;
  Device_Minor = MINOR(file->f_dentry->d_inode->i_rdev);

#ifdef MY_DEBUG
  printk("device_read(%d,%p,%p,%d)\n", Device_Minor, file, buffer, length);
#endif

  /* Number of bytes actually written to the buffer */
  if(length <= 0)
    return -EINVAL;

  if (Device_Minor == STAT_MINOR_NUM) {   // STAT
#ifdef MY_DEBUG
    printk("Read STAT data %d\n",STAT_len);
#endif
      if (length > BUF_LEN) length = BUF_LEN;
      pStatData = &StatData.keylock;
      Message_Ptr = pStatData;      //StatData
      STAT_len = STAT_LEN;
  }
  else if (Device_Minor == KEYBOARD_MINOR_NUM) {   // KEYBOARD
#ifdef MY_DEBUG
    printk("Read KEYBOARD data %d\n",length);
#endif
    if (length > BUF_LEN) length = BUF_LEN;
      pKeyboardData = &KeyboardData[0];
      Message_Ptr = pKeyboardData;
  }
  
  /* If we're at the end of the message, return 0
   * (which signifies end of file) */
  if (Message_Ptr == 0)
    return 0;

  /* Actually put the data into the buffer */
  while (length)  {

    /* Because the buffer is in the user data segment,
     * not the kernel data segment, assignment wouldn't
     * work. Instead, we have to use put_user which
     * copies data from the kernel data segment to the
     * user data segment. */
#ifdef MY_DEBUG
    printk("data %x\n",*Message_Ptr);
#endif
    put_user(*(Message_Ptr++), buffer++);
    length --;
    bytes_read ++;
  }

#ifdef MY_DEBUG
  printk ("Read %d bytes, %d left\n", bytes_read, length);
#endif

  /* ReInitialize the message */

  if (Device_Minor == KEYBOARD_MINOR_NUM) {        // Keyboard
    pKeyboardData = &KeyboardData[0];
    kdata = (unsigned long *)KeyboardData;
  KEYBOARD_STATUS = 0;
  }

  /* Read functions are supposed to return the number
   * of bytes actually inserted into the buffer */
  return bytes_read;
}


/* This function is called when somebody tries to
 * write into our device file. */
static ssize_t device_write(struct file *file,
                            const char *buffer,
                            size_t length,
                            loff_t *offset)
{
  unsigned int i=0;
  Device_Minor = MINOR(file->f_dentry->d_inode->i_rdev);

  /** Write to STAT
  *	send buffer to keyboard firmware
  */
  if(Device_Minor == STAT_MINOR_NUM) {   //STAT
      for(i=0; i<length && i<BUF_LEN; i++) {
        get_user(Message[i], buffer+i);
      }
  }

  /* Write to Keyboard */
  else if(Device_Minor == KEYBOARD_MINOR_NUM) {   //POSKeyboard
    for(i=0; i<length && i<KEYCONF_LEN; i++) {
      get_user(keyconf[i], buffer+i);
#ifdef MY_DEBUG
       printk("Key %x : Konfig %x\n",i,keyconf[i]);
#endif
    }
  }

  /** Write to other devices
  *
  */
  else {
    printk("device_write not available for this device %d\n",Device_Minor);
    i = 0;
  }

  Message_Ptr = Message;

  /* Again, return the number of input characters used */
  return i;
}


/* Timer routine for timeout function */
void jiq_timedout(unsigned long ptr)
{
  // add commands for timeout handling
#ifdef MY_DEBUG
  printk("Timeout\n");
#endif
  PrefixFlag = 0;
    pKeyboardData = &KeyboardData[0];
    kdata = (unsigned long *)KeyboardData;
  KEYBOARD_STATUS = 0;
}

unsigned long checkforposkey(unsigned char key, unsigned char release, unsigned char modifier)
{
  unsigned long retwert;
  unsigned char prefix;
  
  retwert = 0;
  prefix = keyconf[512];
  
  if ((prefix != 0) && (key == prefix))
  { /* prefix empfangen */
#ifdef MY_DEBUG
    printk("Prefix received\n");
#endif
    PrefixFlag = 1;
    return 0;
  }

  if (PrefixFlag) {
#ifdef MY_DEBUG
    printk("Key after Prefix received\n");
#endif
    if (release) PrefixFlag = 0;
    retwert = ((keyconf[512+(key<<1)]) + ((keyconf[513+(key<<1)])<<8)) ;
    if (release) retwert |= 0x80000000ul;
    retwert |= (((unsigned long)modifier)<<16);
  }
  else {
#ifdef MY_DEBUG
    printk("Standard Key received\n");
#endif
    retwert = ((keyconf[key<<1]) + ((keyconf[(key<<1)+1])<<8)) ;
    if (!(retwert)) return retwert;
    if (release) retwert |= 0x80000000ul;
    retwert |= (((unsigned long)modifier)<<16);
  }
  
  return retwert;
}

/* My Keyboard Hook ****************************** */
unsigned char my_kbdhook(unsigned char *kbdata)
{
    unsigned int i;
    unsigned char retwert;
    unsigned char key;
    unsigned long int longkey;

   // retrigger timeout counter
#ifdef MY_DEBUG
    //function is
    printk("timerstrukturadresse:is %x \n",(unsigned int)&jiq_timer);
    printk("modtimer &function is %x \n",(unsigned int)jiq_timer.function);
    printk("modtimer data is %x \n",(unsigned int)jiq_timer.data);
    printk("modtimer expires is %x \n",(unsigned int)jiq_timer.expires);
    mod_timer(&jiq_timer, jiffies + HZ/2);	//retrigger timeout counter 5secs
#else
    mod_timer(&jiq_timer, jiffies + HZ/2);	//retrigger timeout counter 0.5secs
#endif
    retwert = 0;

#ifdef MY_DEBUG
    printk("Keyboarddata");
    for (i=0;i<8;i++)
    {
	printk(" %x",kbdata[i]);
    }
    printk("\n");
#endif

    if (ckbd_u_keyboard)
    { /* keyboard open, use keykonfig */

	memcpy (OldKbdData,NewKbdData,8);
	memcpy (NewKbdData,kbdata,8);  

	for (i = 2; i < 8; i++) {

		if (OldKbdData[i] > 3 && memscan(NewKbdData + 2, OldKbdData[i], 6) == NewKbdData + 8) {
			key = OldKbdData[i];
			longkey = checkforposkey(key,1,NewKbdData[0]);
#ifdef MY_DEBUG
			printk("Key (scancode %x) released.\n", key);
#endif
			if (longkey) {
#ifdef MY_DEBUG
				printk("JPOS event %x reported.\n", (unsigned int)longkey);
#endif
				if (KEYBOARD_STATUS < (BUF_LEN>>2)) { 
					*kdata = longkey;
					kdata++;
					KEYBOARD_STATUS++;
				}
				retwert = 1;
			}
		}

		if (NewKbdData[i] > 3 && memscan(OldKbdData + 2, NewKbdData[i], 6) == OldKbdData + 8) {
			key = NewKbdData[i];
			longkey = checkforposkey(key,0,NewKbdData[0]);
#ifdef MY_DEBUG
			printk("Key (scancode %#x) pressed.\n", key);
#endif
			if (longkey) {
#ifdef MY_DEBUG
				printk("JPOS event %#x reported.\n", (unsigned int)longkey);
#endif
				if (KEYBOARD_STATUS < (BUF_LEN>>2)) { 
					*kdata = longkey;
					kdata++;
					KEYBOARD_STATUS++;
				}
				retwert = 1;
			}
		}
	}
    }
    return retwert;
}


/* This structure will hold the functions to be called
 * when a process does something to the device we
 * created. Since a pointer to this structure is kept in
 * the devices table, it can't be local to
 * init_module. NULL is for unimplemented functions. */
struct file_operations Fops = {
  owner:  THIS_MODULE,
  read:   device_read,
  write:  device_write,
  open:   device_open,
  release: device_release
};


/* Initialize and Register the character device */
int init_ckbd()
{
  int ret_val;
  int i;

  /* Register the character device (atleast try) */
  ret_val = register_chrdev(CKBD_MAJOR_NUM,
                            DEVICE_NAME,
                            &Fops);

  /* Negative values signify an error */
  if (ret_val < 0) {
    printk ("%s failed with %d\n",
            "Sorry, registering the character device ",
            ret_val);
    return ret_val;
  }
  Device_Major = ret_val;

  printk ("chyposusbkbd registered with major device number %d.\n", Device_Major);

  spin_lock_init(&ckbd_u_lock);   //init spin_lock

  /* init Keylock data */
  KEYLOCK_STATUS = 0;

  /* init STAT data */
  StatData.keylock = 0;
  StatData.scancodenr = 0;
  STAT_len = 0;
  pStatData = 0;

  /* init Keyboard data */
  KeyboardData[0] = 0;
  KEYBOARD_STATUS = 0;
  PrefixFlag = 0;
  pKeyboardData = &KeyboardData[0];
  kdata = (unsigned long *)KeyboardData;

  // init timeout counter
  init_timer(&jiq_timer);
  jiq_timer.function = jiq_timedout;
  jiq_timer.data = 0;
#ifdef MY_DEBUG
	printk("timers initialised\n");
#endif
  jiq_timer.expires = jiffies + HZ/50;	// 1 sec
  add_timer(&jiq_timer);
//  del_timer_sync(&jiq_timer);

  for (i=0;i<KEYCONF_LEN;i++)
  {
    keyconf[i]=0;
  }
  return 0;
}


/* Cleanup - unregister the appropriate file from /proc */
void cleanup_ckbd()
{
  // delete timeout counter
  del_timer_sync(&jiq_timer);

  if(Device_Major == 0)
    printk("Error because Device_Major is 0\n");

 unregister_chrdev (Device_Major, DEVICE_NAME);
}
